use utf8;
use strict;
use warnings FATAL => 'all';

use Encode;

use Xyzzy;
use Xyzzy::Header;
use Xyzzy::Request;

sub concat {
	return undef if grep !defined, @_;
	return join('/', @_);
}

my %configs;

$0 = "xyzzy (virgin)";

while(my $req = new Xyzzy::Request) {
	my $oldout = *STDOUT;
	*STDOUT = *STDERR;
	my @output;

	my @path = grep { $_ ne '' } split('/', $req->path_info);

	my ($generator, $status, $content);
	eval {
		die "\$XYZZYRC not set\n"
			unless exists $ENV{XYZZYRC};
		my $config;
		my $configfile = $ENV{XYZZYRC};
		$0 = "xyzzy $configfile";
		if(exists $configs{$configfile}) {
			$config = $configs{$configfile};
		} else {
			$config = new Xyzzy;
			$config->include($configfile);
			$configs{$configfile} = $config;
		}
		$req->config($config);

		$generator = $config->generators;
		while(ref $generator) {
			$generator = $generator->{shift @path or ''}
		}

		$generator = $config->fallback
			unless defined $generator;

		die "No generator found!\n"
			unless defined $generator;

		$generator->generate($req, $config, \@path);
		$status = $req->status;
		$content = $req->content;

		if(defined $content) {
			$content = Encode::encode_utf8($content)
				if Encode::is_utf8($content);
		} else {
			$req->mimetype('text/html', charset => 'UTF-8');
			$content = $status->html;
		}
		$req->setheader('Content-Length', length($content));

		push @output, "Status: $status\r\n";
		my $headers = $req->outheaders;
		while(my ($key, $vals) = each(%$headers)) {
			my $concat = 1;
			if(@$vals > 1) {
				foreach my $val (@$vals) {
					$concat = 0
						unless Xyzzy::Header::istoken($val);
				}
			}
			if($concat) {
				push @output, $key, ': ', join(', ', @$vals), "\r\n";
			} else {
				foreach my $val (@$vals) {
					push @output, $key, ': ', $val, "\r\n";
				}
			}
		}
		push @output, "\r\n", $content;
	};
	$0 = "xyzzy (output)";
	if($@) {
		warn $@;
		$status = new Xyzzy::Status(500);
		@output = $status->response;
	}

	foreach(@output) {
		syswrite($oldout, $_)
			or die "Writing output: $!\n";
	}
	*STDOUT = $oldout;
	$0 = "xyzzy (idle)";
}
